/*
 * Input driver for Microsoft MCE Remote Control - originally for 2005 keyboard and mouse
 *
 * (C) by Florian Demski
 * Updated 19th Jan 2010
 *
 * Based upon lirc_mcesub2.c from Martin A. Blatter
 * with MCE gen 1 init code taken from the lirc project driver
 * Updated January 2010 by Graham Auld <mceusb@graham.auld.me.uk>
 *
 **********************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
	#error "This modules needs kernel 2.6.10 or higher"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	#include <linux/compat.h>
#else
	#include <linux/config.h>
#endif
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/input.h>
/*#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
Centos 5.4, 2.6.18.9.1 kernel actually has usb/input rather than usb_input
G.Auld	6/Jan/2010
*/
	#include <linux/usb/input.h>
/*#else
	#include <linux/usb_input.h>
#endif*/

#define DRIVER_VERSION		"0.1.5"
#define DRIVER_AUTHOR  		"Florian Demski & Graham Auld"
#define DRIVER_DESC    		"Input driver for Microsoft MCE Remote Control"
#define DRIVER_NAME    		"mod_mce"

/* repeats as number of ir signals to take before repeating*/
#define	SOFTWARE_REPEAT
#define	FAST_REPEAT		0
#define	SLOW_REPEAT		7

#define NAME_BUFSIZE		80
#define USB_BUFLEN		16
#define PEAK_BUFLEN	        10000

#define MCE_GAP         	2000

#define MCE_PACKET_HEADER1 	0x84
#define MCE_PACKET_HEADER2 	0x83
#define MCE_PACKET_HEADER3 	0x82
#define MCE_PACKET_HEADER4 	0x81

#define SYNC_MIN 	52
#define SYNC_MAX 	66
#define RC6_1_MIN 	6	//7
#define RC6_1_MAX 	13	//12
#define RC6_2_MIN 	15	//15
#define RC6_2_MAX 	20	//20
#define RC6_3_MIN 	24	//24
#define RC6_3_MAX 	32	//32

#define RC5_1_MIN	3
#define RC5_1_MAX 	9
#define RC5_2_MIN 	10
#define RC5_2_MAX 	15

#define RC5_KEY_BUFLEN		74
#define RC5_MOUSE_BUFLEN 	68

#define RC5_KEY_START		42
#define RC5_MASK_START		58

#define MOUSE_Y_START       	26
#define MOUSE_X_START		40
#define MOUSE_BTN_START		54
#define MOUSE_DATA_END		58

#define MASK_KEYS			0xE0

/* general constants */
#define SUCCESS			0
#define SEND_FLAG_COMPLETE	2
#define RECV_FLAG_IN_PROGRESS	3

#define PULSE 	2
#define SPACE 	1

#define PHILUSB_INBOUND		1
#define PHILUSB_OUTBOUND	2

#define VENDOR_MS		0x045e
#define VENDOR_PHILIPS		0x0471
#define VENDOR_SMK		0x0609
#define VENDOR_TATUNG		0x1460
#define VENDOR_GATEWAY		0x107b
#define VENDOR_SHUTTLE		0x1308

#define dprintk(fmt, args...) if(debug) printk(KERN_DEBUG fmt, ## args);

	//they KEY_defines come from input.h
	//probably at /usr/include/linux/input.h if you have no kernel sources installed
#define	SHIFT		0x100
#define	CTRL		0x200
#define	ALT		0x400

//some defaults suitable for mythtv

#define	MCE_CLEAR	KEY_D //KEY_BACKSPACE
#define MCE_PCPOWER	CTRL+ALT+KEY_BACKSPACE //KEY_POWER
#define MCE_GREENBUTTON	CTRL+ALT+KEY_M	//jumppoint MainMenu
#define MCE_MUTE	KEY_MUTE //KEY_F9 //KEY_MUTE
#define MCE_INFO	KEY_I
#define MCE_VOLUP	KEY_VOLUMEUP //KEY_F11	//KEY_VOLUMEUP
#define MCE_VOLDN	KEY_VOLUMEDOWN //KEY_F10	//KEY_VOLUMEDOWN
#define MCE_CHUP	KEY_SLASH //skip to next favourite channel
#define MCE_CHDN	KEY_H //go to previous channel
#define MCE_FWD		SHIFT+KEY_DOT
#define MCE_REW		SHIFT+KEY_COMMA
#define MCE_PLAY	KEY_SPACE
#define MCE_RECORD	KEY_R
#define MCE_PAUSE	KEY_P
#define MCE_STOP	KEY_ESC
#define MCE_SKIP	KEY_PAGEDOWN
#define MCE_REPLAY	KEY_PAGEUP
#define MCE_HASH	KEY_BACKSPACE //come in handy //KEY_BACKSLASH //is actually '#'
#define MCE_STAR	SHIFT+KEY_SLASH //key '?' - mark channel as favourite
#define MCE_OK		KEY_ENTER
#define MCE_BACK	KEY_ESC
#define MCE_DVDMENU	KEY_M
#define MCE_LIVETV	CTRL+ALT+KEY_T
#define MCE_GUIDE	CTRL+ALT+KEY_G
#define MCE_RECTV	CTRL+ALT+KEY_R
#define MCE_TEXT	KEY_F7
#define MCE_RED		KEY_F2
#define MCE_GREEN	KEY_F3
#define MCE_YELLOW	KEY_F4
#define MCE_BLUE	KEY_F5



static int debug = 0;


static unsigned short mce_rc6_remote_keycode[256] = {
	KEY_0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6, KEY_7, KEY_8, //0x00
	KEY_9, MCE_CLEAR, KEY_ENTER, MCE_PCPOWER, MCE_GREENBUTTON, MCE_MUTE, MCE_INFO, //0x08
	MCE_VOLUP,  MCE_VOLDN,  MCE_CHUP,  MCE_CHDN,  MCE_FWD,  MCE_REW,  MCE_PLAY,  MCE_RECORD,  MCE_PAUSE, //0x10
	MCE_STOP,  MCE_SKIP,  MCE_REPLAY,  MCE_HASH,  MCE_STAR,  KEY_UP,  KEY_DOWN, //0x18
	KEY_LEFT, KEY_RIGHT,  MCE_OK,  MCE_BACK,  MCE_DVDMENU,  MCE_LIVETV,  MCE_GUIDE,  0,  0, //20
	0,  0,  0,  0,  0,  0,  0, //0x28
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0x30
        0,  0,  0,  0,  0,  0,  0,  0,  MCE_RECTV,  0,  0,  0,  0,  0,  0,  0, //0x40
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  MCE_TEXT,  MCE_RED, MCE_GREEN,  MCE_YELLOW,  MCE_BLUE,  0, //0x50
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0x60
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0x70
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0x80
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0x90
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0xa0
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0xb0
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0xc0
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0xd0
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //0xe0
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  KEY_LEFTSHIFT,  KEY_LEFTCTRL,  KEY_LEFTALT, //0xf0
};


static struct usb_device_id usb_table [] = {
	{ USB_DEVICE(VENDOR_MS,		0x006d) },	/* MS MCEUSB box */
	{ USB_DEVICE(VENDOR_PHILIPS, 	0x0815) },	/* Philips eHome Infrared Transciever */
	{ USB_DEVICE(VENDOR_SMK, 	0x031d) },	/* SMK/Toshiba G83C0004D410 */
	{ USB_DEVICE(VENDOR_TATUNG, 	0x9150) },  /* Tatung eHome Infrared Transceiver */
	{ USB_DEVICE(VENDOR_SHUTTLE, 	0xc001) }, 	/* Shuttle eHome Infrared Transceiver */
 	{ USB_DEVICE(VENDOR_GATEWAY, 	0x3009) },	/* Gateway eHome Infrared Transceiver */
	{ }											/* Terminating entry */
};

static struct usb_device_id microsoft_gen1_list[] = {
 	{ USB_DEVICE(VENDOR_MS,		0x006d) },
 	{}
 };


typedef struct {
	int pulse;
	int duration;
} peak;


/* data structure for each usb remote */
struct irctl {
	/* usb */
	struct input_dev *idev;
	struct usb_device *usbdev;
	struct urb *urb_in;
	int devnum;
//	unsigned char pressed[256/8];
	struct usb_endpoint_descriptor *usb_ep_in;
	struct {
  		u32 microsoft_gen1:1;
		u32 toggle_bit:1;
  		u32 reserved:30;
 	} flags;

	unsigned char	gen1_counter;
	/* buffers and dma */
	unsigned char *buf_in;
	peak peaks[PEAK_BUFLEN];
	int peak_index;
	
	int in_sync;
	int sync_pos;
	
	unsigned int len_in;
	dma_addr_t dma_in;
	
	int last_space;
  	int send_flags;
  	
  	char name[NAME_BUFSIZE];
  	char phys[NAME_BUFSIZE];

	int prev_key;
	int repeat_counter;
};

/* init strings */
static char init1[] = {0x00, 0xff, 0xaa, 0xff, 0x0b};
static char init2[] = {0xff, 0x18};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void usb_async_callback(struct urb *urb, struct pt_regs *regs)
#else
static void usb_async_callback(struct urb *urb)
#endif
{
	struct irctl *ir;
	if (!urb)
		return;
	
	if ((ir = urb->context)) {
		dprintk(DRIVER_NAME "[%d]: callback called (status=%d len=%d)\n",ir->devnum, urb->status, urb->actual_length);
	}
}


/* request incoming or send outgoing usb packet - used to initialize remote */
static void request_packet_async(struct irctl *ir, struct usb_endpoint_descriptor *ep, unsigned char* data, int size, int urb_type)
{
	int res;
	struct urb *async_urb;
	unsigned char *async_buf;

	if (urb_type) {
    		if ((async_urb = usb_alloc_urb(0, GFP_KERNEL))) {
			/* alloc buffer */
			if ((async_buf = kmalloc(size, GFP_KERNEL))) {
				if (urb_type==PHILUSB_OUTBOUND) {
					/* outbound data */
					usb_fill_int_urb(async_urb, ir->usbdev, usb_sndintpipe(ir->usbdev, ep->bEndpointAddress),
					 async_buf, size, usb_async_callback, ir, ep->bInterval);
					memcpy(async_buf, data, size);
				} else {
					/* inbound data */
					usb_fill_int_urb(async_urb, ir->usbdev, usb_rcvintpipe(ir->usbdev, ep->bEndpointAddress),
					 async_buf, size, usb_async_callback, ir, ep->bInterval);
				}
			} else {
				usb_free_urb(async_urb);
				return;
			}
		}
	} else {
		/* standard request */
		async_urb=ir->urb_in;
		ir->send_flags = RECV_FLAG_IN_PROGRESS;
	}

	dprintk(DRIVER_NAME "[%d]: receive request called (size=%#x)\n", ir->devnum, size);
	async_urb->transfer_buffer_length = size;
	async_urb->dev = ir->usbdev;
	if ((res=usb_submit_urb(async_urb, GFP_ATOMIC))) {
	    	dprintk(DRIVER_NAME "[%d]: receive request FAILED! (res=%d)\n", ir->devnum, res);
	    	return;
	}
	dprintk(DRIVER_NAME "[%d]: receive request complete (res=%d)\n", ir->devnum, res);
}

static void do_rc6_remote(struct irctl *ir, peak *peaks, int num_peaks)
{
//so each peak contains pulse and duration.
	int i;
	int j;
	int k;
	int step1[1001];
	int step2[500];
	unsigned short tc;
	int new_key;

	dprintk("ok, got %d peaks (marks and spaces of multiple lengths)\n",num_peaks);
//1 start by constructing the bitstream
//2 test each peak to determine length 1,2,3
//3 then add appropriate number of 1's or 0's to buffer
//4 then add a zero on the end (possibly missing)
//5 check the RC6 header and de-manchester the bitstream
//6 take button byte and lookup keypress

        for (i=0, j=0; i<num_peaks; i++) {
                if (RC6_1_MIN <= peaks[i].duration && peaks[i].duration <= RC6_1_MAX) {
                        step1[j++] = peaks[i].pulse;
                        if (j == 1000) return;
                } else if (RC6_2_MIN <= peaks[i].duration && peaks[i].duration <= RC6_2_MAX) {
                        step1[j++] = peaks[i].pulse;
                        if (j == 1000) return;
                        step1[j++] = peaks[i].pulse;
                        if (j == 1000) return;
                } else if (RC6_3_MIN <= peaks[i].duration && peaks[i].duration <= RC6_3_MAX) {
                        step1[j++] = peaks[i].pulse;
                        if (j == 1000) return;
                        step1[j++] = peaks[i].pulse;
                        if (j == 1000) return;
                        step1[j++] = peaks[i].pulse;
                        if (j == 1000) return;
                }

        }
//step 1, 2 and 3 complete.
//step 4

	dprintk("checking RC6 header\n");
//check out.
//should start with 00 10 10 10 01 sync
//then followed by 00 11 toggle bit (11 00) for tog...
//except, 1 is zero and 2 is 1
	if (step1[0]-1 || step1[1]-1) return;
	if (step1[2]==1 || step1[3]-1) return;
	if (step1[4]==1 || step1[5]-1) return;
	if (step1[6]==1 || step1[7]-1) return;
	if (step1[8]-1 || step1[9]==1) return;
//togglecheck
/*
	if (step1[10]-1 && step1[11]-1 && step1[12]==1 && step1[13]==1) {
		ir->flags.toggle_bit = true;
	dprintk("toggle true\n");
	} else {
		ir->flags.toggle_bit = false;
	dprintk("toggle false\n");
	}
	*/
	
	dprintk("RC6 header found, bitstream is %d bits (%d after header) (%d after demanchestering\n",j,j-14,(j-14)/2);
	//bitstream of around 79bits for one button push

//read step1 from after RC6 header and de-manchester outputting into step2
	for (i=14, k=0; i<j; i+=2) {
		if (step1[i]==1 && step1[i+1]==2) {
			step2[k++]=0;
		} else if (step1[i]==2 && step1[i+1]==1) {
			step2[k++]=1;
		} else {
			if (i==76) {
				step2[k++]=step1[i]-1;
				//if bit 76 (second last) then make up the second manchester bit as it is sometimes missing
			} else {
				dprintk("manchester violation at bits %d&%d of %d! They are %d and %d\n",i,i+1,j,step1[i],step1[i+1]);
				return;	//abandon if manchester is violated
			}
		}
	}

	if (step2[16]) {
        	ir->flags.toggle_bit = true;
	        dprintk("toggle true\n");
        } else {
                ir->flags.toggle_bit = false;
        	dprintk("toggle false\n");
        }


	dprintk("demanchesterd and now have %d bits\n",k);
	if (k<32) return; //need 32bits to have a packet

	//step2 now containins the bitstream as binary
	/*for(i=0;i<32;i+=4) {
		dprintk("codes: %d%d%d%d\n",step2[i],step2[i+1],step2[i+2],step2[i+3]);
	}*/

	//looks at step2 binary code and converts to int
	//	24 25 26 27   28 29 30 31
	k=0;
	j=1;
	for(i=0;i<8;i++,j*=2){
		if (step2[31-i])
		k+=j;
	}
	dprintk("IR button k is %d\n",k);

#ifdef SOFTWARE_REPEAT	
	new_key=k;
	if (ir->flags.toggle_bit) new_key+=1024;

	tc = mce_rc6_remote_keycode[k];
        dprintk("key from remote was %d, table gave value %d\n",k,tc);
	if (tc & 0x100) dprintk("shifting key %d\n",tc&0xFF);

	if (new_key==ir->prev_key) {
		ir->repeat_counter--;
		if (ir->repeat_counter <=0) {
			ir->repeat_counter=FAST_REPEAT;

			//tc = mce_rc6_remote_keycode[k];
			//dprintk("key from remote was %d, table gave value %d\n",k,tc);
			if (tc & 0x100) input_report_key(ir->idev, KEY_LEFTSHIFT, 1);
			if (tc & 0x200) input_report_key(ir->idev, KEY_LEFTCTRL, 1);
			if (tc & 0x400) input_report_key(ir->idev, KEY_LEFTALT, 1);
		        input_report_key(ir->idev, tc&0xFF, 1);//keypress
		        input_report_key(ir->idev, tc&0xFF, 0);//keyrelease
			if (tc & 0x100) input_report_key(ir->idev, KEY_LEFTSHIFT, 0);
                        if (tc & 0x200) input_report_key(ir->idev, KEY_LEFTCTRL, 0);
                        if (tc & 0x400) input_report_key(ir->idev, KEY_LEFTALT, 0);


		}
	} else {
		//tc = mce_rc6_remote_keycode[k];
		//dprintk("key from remote was %d, table gave value %d\n",k,tc);
		if (tc & 0x100) input_report_key(ir->idev, KEY_LEFTSHIFT, 1);
                if (tc & 0x200) input_report_key(ir->idev, KEY_LEFTCTRL, 1);
                if (tc & 0x400) input_report_key(ir->idev, KEY_LEFTALT, 1);
        	input_report_key(ir->idev, tc&0xFF, 1);//keypress
	        input_report_key(ir->idev, tc&0xFF, 0);//keyrelease
		if (tc & 0x100) input_report_key(ir->idev, KEY_LEFTSHIFT, 0);
                if (tc & 0x200) input_report_key(ir->idev, KEY_LEFTCTRL, 0);
                if (tc & 0x400) input_report_key(ir->idev, KEY_LEFTALT, 0);

		ir->prev_key=new_key;
		ir->repeat_counter=SLOW_REPEAT;
	}
#endif
#ifndef	SOFTWARE_REPEAT
//IR dosn't give us key up codes (well not the RC6 MCE remotes anyway)
//what we do get is a bit toggle on re press
//this means it's a bit tricky to do key press/release detection to allow
//the input subsystem to do key autorepeat.

	tc = mce_rc6_remote_keycode[k];
	input_report_key(ir->idev, tc, 1);//keypress
	input_report_key(ir->idev, tc, 0);//keyrelease

#endif

}


static void do_analyze(struct irctl *ir, peak *peaks, int num_peaks)
{
	do_rc6_remote(ir, peaks, num_peaks);
	//could look for RC5 next if you like
	//I removed the RC5 keyboard and mouse code as it caused kernel fault and I have no hardware to debug it.
}


static void decode_buffer(struct irctl *ir, int len)
{
	int i;
	int pulse, length, sync_pos;
	int start_index = 0;
	
	sync_pos = -1;

	if (len <= 0) return;
	if (ir->flags.microsoft_gen1 && len <=2) return;

	if (ir->flags.microsoft_gen1)
		start_index = 2;

	dprintk("Buffer is %d long with a gen1counter of %d\n",len,ir->gen1_counter);
	if (((ir->buf_in[start_index] >= MCE_PACKET_HEADER4) && (ir->buf_in[start_index] <= MCE_PACKET_HEADER1)) || (ir->flags.microsoft_gen1 && (ir->gen1_counter > 0) )) {//only process where the first buffer part is a header
	//or a gen1 device still expecting data.
		//then gen 1s blat stuff out all over the show, they do it with a header then 4 bytes but spread over packets.
		for (i=start_index; i<len; i++) {
			if ((ir->buf_in[i] >= MCE_PACKET_HEADER4) && (ir->buf_in[i] <= MCE_PACKET_HEADER1)) { //ignore anything 81-84
				//if gen1 then preset counter
				if (ir->flags.microsoft_gen1) {
		                	ir->gen1_counter = 4;
					dprintk("start new 4bytepack at urb index %d\n",i);
				}
					
				continue;
			} else {
				if (ir->flags.microsoft_gen1)
                                        ir->gen1_counter--;
			}
			


			pulse  = (ir->buf_in[i] & 0x80) ? PULSE : SPACE;
			length = (ir->buf_in[i] & 0x7F);
			
			if (ir->peaks[ir->peak_index].pulse != pulse) { //when it's not a pulse anymore look at the lengh
				if ((ir->peaks[ir->peak_index].pulse == PULSE) && 
					(SYNC_MIN <= ir->peaks[ir->peak_index].duration) &&
					(ir->peaks[ir->peak_index].duration <= SYNC_MAX)) {
						dprintk("found valid sync\n");
						if (ir->in_sync) {
							do_analyze(ir, &ir->peaks[ir->sync_pos], ir->peak_index - ir->sync_pos + 1);
						}
						ir->in_sync = 1;
						ir->sync_pos = 0;
						ir->peak_index = 0;
						
				}
				
				if (ir->peaks[ir->peak_index].duration == MCE_GAP) {
					dprintk("found mce gap\n");
					if (ir->in_sync) {
						do_analyze(ir, &ir->peaks[ir->sync_pos], ir->peak_index - ir->sync_pos + 1);
					}
					ir->in_sync = 0;
					ir->peak_index = 0;
		            		ir->peaks[0].pulse=0;
		            		break;
				}
				ir->peak_index = (++ir->peak_index) % PEAK_BUFLEN;
				if (ir->peak_index == 0) dprintk(DRIVER_NAME ": buffer overflow\n");
				ir->peaks[ir->peak_index].pulse = pulse;
				ir->peaks[ir->peak_index].duration = length;
			} else {//when it is a pulse, increment the duration
				ir->peaks[ir->peak_index].duration += length;
			}
		}
	}
}



#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void usb_recv(struct urb *urb, struct pt_regs *regs)
#else
static void usb_recv(struct urb *urb)
#endif
{
	struct irctl *ir;
	int len;

	if (!urb)
		return;

	if (!(ir = urb->context)) {
		usb_unlink_urb(urb);
		return;
	}

	len = urb->actual_length;

	if (ir->send_flags==RECV_FLAG_IN_PROGRESS) {
	  	ir->send_flags = SEND_FLAG_COMPLETE;
		dprintk(DRIVER_NAME "[%d]: setup answer received %d bytes\n",ir->devnum,len);
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
        input_regs(ir->idev, regs);
#endif

	switch (urb->status) {
	/* success */
	case 0:
		decode_buffer(ir, len);
        	input_sync(ir->idev);
		break;

	/* unlink */
	case -ECONNRESET:
	case -ENOENT:
	case -ESHUTDOWN:
		usb_unlink_urb(urb);
		return;

	case -EPIPE:
	default:
		break;
	}

	/* resubmit urb */
	usb_submit_urb(urb, GFP_ATOMIC);//request for async transfer for an emdpoint (urb)
}

static int input_open(struct input_dev *id)
{
	return 0;
}

static void input_close(struct input_dev *id)
{
}

static int mceusb_gen1_init(struct irctl *ir)
{
	int i, ret;
	char junk[64], data[8];
	int partial = 0;

	/*
	 * Clear off the first few messages. These look like calibration
	 * or test data, I can't really tell. This also flushes in case
	 * we have random ir data queued up.
	 */
	for (i = 0; i < 40; i++)
		usb_bulk_msg(ir->usbdev,
			usb_rcvbulkpipe(ir->usbdev,
				ir->usb_ep_in->bEndpointAddress),
			junk, 64, &partial, HZ * 10);

//	ir->is_pulse = 1;
 
 	memset(data, 0, 8);

	/* Get Status */
	ret = usb_control_msg(ir->usbdev, usb_rcvctrlpipe(ir->usbdev, 0),
			      USB_REQ_GET_STATUS, USB_DIR_IN,
			      0, 0, data, 2, HZ * 3);

	/*    ret = usb_get_status( ir->usbdev, 0, 0, data ); */
	dprintk("%s - ret = %d status = 0x%x 0x%x\n", __func__,
		ret, data[0], data[1]);

	/*
	 * This is a strange one. They issue a set address to the device
 	 * on the receive control pipe and expect a certain value pair back
	 */
	memset(data, 0, 8);

	ret = usb_control_msg(ir->usbdev, usb_rcvctrlpipe(ir->usbdev, 0),
			      USB_REQ_SET_ADDRESS, USB_TYPE_VENDOR, 0, 0,
			      data, 2, HZ * 3);
	dprintk("%s - ret = %d, devnum = %d\n",
		__func__, ret, ir->usbdev->devnum);
	dprintk("%s - data[0] = %d, data[1] = %d\n",
		__func__, data[0], data[1]);

	/* set feature: bit rate 38400 bps */
	ret = usb_control_msg(ir->usbdev, usb_sndctrlpipe(ir->usbdev, 0),
			      USB_REQ_SET_FEATURE, USB_TYPE_VENDOR,
			      0xc04e, 0x0000, NULL, 0, HZ * 3);

	dprintk("%s - ret = %d\n", __func__, ret);

	/* bRequest 4: set char length to 8 bits */
	ret = usb_control_msg(ir->usbdev, usb_sndctrlpipe(ir->usbdev, 0),
			      4, USB_TYPE_VENDOR,
			      0x0808, 0x0000, NULL, 0, HZ * 3);
	dprintk("%s - retB = %d\n", __func__, ret);

	/* bRequest 2: set handshaking to use DTR/DSR */
	ret = usb_control_msg(ir->usbdev, usb_sndctrlpipe(ir->usbdev, 0),
			      2, USB_TYPE_VENDOR,
			      0x0000, 0x0100, NULL, 0, HZ * 3);
	dprintk("%s - retC = %d\n", __func__, ret);

	return ret;
 
};


static int usb_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct usb_host_interface *idesc;
	struct usb_endpoint_descriptor *ep=NULL, *ep_in=NULL, *ep_out=NULL;
	struct usb_host_config *config;
	struct input_dev *input_dev;
	struct irctl *ir = NULL;
	int devnum, pipe, maxp;
	int i;
	char buf[63], name[128]="";
	int mem_failure = 0;
	int is_microsoft_gen1;


	dprintk(DRIVER_NAME ": usb probe called\n");

	usb_reset_device(dev);

	config=dev->actconfig;

	idesc = intf->cur_altsetting;

	is_microsoft_gen1 = usb_match_id(intf, microsoft_gen1_list) ? 1 : 0;


	/* step through the endpoints to find first bulk in and out endpoint */
	for (i = 0; i < idesc->desc.bNumEndpoints; ++i) {
		ep = &idesc->endpoint[i].desc;

		if ((ep_in == NULL)
			&& ((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
			&& (((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)
			||  ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_INT))) {

			dprintk(DRIVER_NAME ": acceptable inbound endpoint found\n");
			ep_in = ep;
			ep_in->bmAttributes=USB_ENDPOINT_XFER_INT;
			ep_in->bInterval=1;
		}

		if ((ep_out == NULL)
			&& ((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_OUT)
			&& (((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK)
			||  ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_INT))) {

			dprintk(DRIVER_NAME ": acceptable outbound endpoint found\n");
			ep_out = ep;
			ep_out->bmAttributes=USB_ENDPOINT_XFER_INT;
			ep_out->bInterval=1;
		}
	}
	if (ep_in == NULL) {
		dprintk(DRIVER_NAME ": inbound and/or endpoint not found\n");
		return -ENODEV;
	}

	input_dev = input_allocate_device();
	if (input_dev == NULL) {
		dprintk(DRIVER_NAME ": could not get input device\n");
		return -ENODEV;
	}

	devnum = dev->devnum;
	pipe = usb_rcvintpipe(dev, ep_in->bEndpointAddress);
	maxp = usb_maxpacket(dev, pipe, usb_pipeout(pipe));

	/* allocate kernel memory */
	mem_failure = 0;
	if (!(ir = kmalloc(sizeof(struct irctl), GFP_KERNEL))) {
		mem_failure = 1;
	} else {
		memset(ir, 0, sizeof(struct irctl)); /* that should set peak_index to 0 */

	    if (!(ir->buf_in = usb_buffer_alloc(dev, maxp, GFP_ATOMIC, &ir->dma_in))) {
			mem_failure = 5;
		} else if (!(ir->urb_in = usb_alloc_urb(0, GFP_KERNEL))) {
			mem_failure = 7;
		}
	}

	/* free allocated memory incase of failure */
	switch (mem_failure) {
	case 7:
		usb_free_urb(ir->urb_in);
	case 5:
		usb_buffer_free(dev, maxp, ir->buf_in, ir->dma_in);
		kfree(ir);
	case 1:
		input_free_device(input_dev);
		printk(DRIVER_NAME "[%d]: out of memory (code=%d)\n",
			devnum, mem_failure);
		return -ENOMEM;
	}

	usb_make_path(dev, ir->phys, sizeof(ir->phys));
	strlcpy(ir->phys, "/input0", sizeof(ir->phys));
	if (dev->manufacturer)
		strlcpy(ir->name, dev->manufacturer, sizeof(ir->name));
	if (dev->product)
		snprintf(ir->name, sizeof(ir->name), "%s %s", ir->name, dev->product);
	if (!strlen(ir->name))
		snprintf(ir->name, sizeof(ir->name), DRIVER_DESC "(%04x,%04x)",
			dev->descriptor.idVendor, dev->descriptor.idProduct);
			
	input_dev->open = input_open;
	input_dev->close = input_close;
	input_dev->name = ir->name;
	input_dev->phys = ir->phys;
	usb_to_input_id(dev, &input_dev->id);
	input_dev->cdev.dev = &dev->dev;
   	//input_dev->evbit[0] = BIT(EV_KEY);//only be generating key presses | BIT(EV_REL);
	input_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_REP);
	//enable key autorepeat
	//input_dev->evbit[1] = BIT(EV_REP);

   	//input_dev->relbit[0] = BIT(REL_X) | BIT(REL_Y);
	for (i=1; i<256; i++) {
		//keypresses as listed in this array
		//now any key from 0-255 can be sent
		set_bit(i, input_dev->keybit);
	} 
 	
    clear_bit(0, input_dev->keybit);	
  // 	input_dev->keybit[LONG(BTN_MOUSE)] = BIT(BTN_LEFT) | BIT(BTN_RIGHT);

	input_register_device(input_dev);
	
//	memset(ir->pressed, 0, sizeof(ir->pressed));

    ir->idev = input_dev;
	ir->devnum = devnum;
	ir->usbdev = dev;
	ir->len_in = maxp;

	if (dev->descriptor.iManufacturer
		&& usb_string(dev, dev->descriptor.iManufacturer, buf, 63) > 0)
		strncpy(name, buf, 128);
	if (dev->descriptor.iProduct
		&& usb_string(dev, dev->descriptor.iProduct, buf, 63) > 0)
		snprintf(name, 128, "%s %s", name, buf);
	printk(DRIVER_NAME "[%d]: %s on usb%d:%d\n", devnum, name,
	       dev->bus->busnum, devnum);

	/* inbound data */
	usb_fill_int_urb(ir->urb_in, dev, pipe, ir->buf_in,
		maxp, usb_recv, ir, ep_in->bInterval);


	if (is_microsoft_gen1) {
		//intialise MCEUSB gen1 device
		dprintk("mcegen1\n");
		ir->flags.microsoft_gen1 = 1;
		ir->usb_ep_in = ep_in;
		mceusb_gen1_init(ir);
	} else {
		dprintk("not gen1 mcen\n");
		/* initialize device */
		request_packet_async( ir, ep_in, NULL, maxp, PHILUSB_INBOUND );
		request_packet_async( ir, ep_in, NULL, maxp, PHILUSB_INBOUND );
		request_packet_async( ir, ep_out, init1, sizeof(init1), PHILUSB_OUTBOUND );
		request_packet_async( ir, ep_in, NULL, maxp, PHILUSB_INBOUND );
		request_packet_async( ir, ep_out, init2, sizeof(init2), PHILUSB_OUTBOUND );
	}

	request_packet_async( ir, ep_in, NULL, maxp, 0);

	usb_set_intfdata(intf, ir);
	return 0;
}


static void usb_disconnect(struct usb_interface *intf)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct irctl *ir = usb_get_intfdata(intf);

	usb_set_intfdata(intf, NULL);

	if (!ir)
		return;

	if (ir->idev) {
		input_unregister_device(ir->idev);
		ir->idev = NULL;
	}
	ir->usbdev = NULL;

	usb_kill_urb(ir->urb_in);
	usb_free_urb(ir->urb_in);
	usb_buffer_free(dev, ir->len_in, ir->buf_in, ir->dma_in);
}

static struct usb_driver usb_driver = {
	.name = DRIVER_NAME,
	.probe = usb_probe,
	.disconnect = usb_disconnect,
	.id_table = usb_table
};

static int __init mod_mce_init(void)
{
	int i;

	printk("\n" DRIVER_NAME ": " DRIVER_DESC " v" DRIVER_VERSION "\n");
	dprintk(DRIVER_NAME " : debug mode enabled\n");
		
	if ((i = usb_register(&usb_driver)) < 0) {
		printk(DRIVER_NAME ": usb register failed, resulst = %d\n", i);
		return -ENODEV;
	}
	return 0;
}

static void __exit mod_mce_exit(void)
{
	printk(DRIVER_NAME ": exiting\n");
	usb_deregister(&usb_driver);
}

module_init(mod_mce_init);
module_exit(mod_mce_exit);

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE("GPL");

module_param(debug, bool, 0644);
MODULE_PARM_DESC(debug, "Debug enabled or not");

MODULE_ALIAS("usb:v045Ep006Dd*dc*dsc*dp*ic*isc*ip*");
MODULE_ALIAS("usb:v0471p0815d*dc*dsc*dp*ic*isc*ip*");



